@@ -22,10 +22,11 @@ except ImportError:
     _has_pytz = False
 
 
-class _VisibleDeprecationTestCase(object):
+class _DeprecationTestCase(object):
     # Just as warning: warnings uses re.match, so the start of this message
     # must match.
     message = ''
+    warning_cls = DeprecationWarning
 
     def setUp(self):
         self.warn_ctx = warnings.catch_warnings(record=True)
@@ -37,119 +38,17 @@ class _VisibleDeprecationTestCase(object):
         # try to keep the tests cleanly giving only the right warning type.
         # (While checking them set to "error" those are ignored anyway)
         # We still have them show up, because otherwise they would be raised
-        warnings.filterwarnings("always", category=np.VisibleDeprecationWarning)
+        warnings.filterwarnings("always", category=self.warning_cls)
         warnings.filterwarnings("always", message=self.message,
-                                category=np.VisibleDeprecationWarning)
+                                category=self.warning_cls)
 
     def tearDown(self):
         self.warn_ctx.__exit__()
 
     def assert_deprecated(self, function, num=1, ignore_others=False,
                           function_fails=False,
-                          exceptions=(np.VisibleDeprecationWarning,),
+                          exceptions=np._NoValue,
                           args=(), kwargs={}):
-        """Test if VisibleDeprecationWarnings are given and raised.
-
-        This first checks if the function when called gives `num`
-        VisibleDeprecationWarnings, after that it tries to raise these
-        VisibleDeprecationWarnings and compares them with `exceptions`.
-        The exceptions can be different for cases where this code path
-        is simply not anticipated and the exception is replaced.
-
-        Parameters
-        ----------
-        function : callable
-            The function to test
-        num : int
-            Number of VisibleDeprecationWarnings to expect. This should
-            normally be 1.
-        ignore_others : bool
-            Whether warnings of the wrong type should be ignored (note that
-            the message is not checked)
-        function_fails : bool
-            If the function would normally fail, setting this will check for
-            warnings inside a try/except block.
-        exceptions : Exception or tuple of Exceptions
-            Exception to expect when turning the warnings into an error.
-            The default checks for DeprecationWarnings. If exceptions is
-            empty the function is expected to run successfully.
-        args : tuple
-            Arguments for `function`
-        kwargs : dict
-            Keyword arguments for `function`
-        """
-        # reset the log
-        self.log[:] = []
-
-        try:
-            function(*args, **kwargs)
-        except (Exception if function_fails else tuple()):
-            pass
-
-        # just in case, clear the registry
-        num_found = 0
-        for warning in self.log:
-            if warning.category is np.VisibleDeprecationWarning:
-                num_found += 1
-            elif not ignore_others:
-                raise AssertionError(
-                        "expected DeprecationWarning but got: %s" %
-                        (warning.category,))
-        if num is not None and num_found != num:
-            msg = "%i warnings found but %i expected." % (len(self.log), num)
-            lst = [w.category for w in self.log]
-            raise AssertionError("\n".join([msg] + lst))
-
-        with warnings.catch_warnings():
-            warnings.filterwarnings("error", message=self.message,
-                                    category=np.VisibleDeprecationWarning)
-            try:
-                function(*args, **kwargs)
-                if exceptions != tuple():
-                    raise AssertionError(
-                            "No error raised during function call")
-            except exceptions:
-                if exceptions == tuple():
-                    raise AssertionError(
-                            "Error raised during function call")
-
-    def assert_not_deprecated(self, function, args=(), kwargs={}):
-        """Test if VisibleDeprecationWarnings are given and raised.
-
-        This is just a shorthand for:
-
-        self.assert_deprecated(function, num=0, ignore_others=True,
-                        exceptions=tuple(), args=args, kwargs=kwargs)
-        """
-        self.assert_deprecated(function, num=0, ignore_others=True,
-                        exceptions=tuple(), args=args, kwargs=kwargs)
-
-
-class _DeprecationTestCase(object):
-    # Just as warning: warnings uses re.match, so the start of this message
-    # must match.
-    message = ''
-
-    def setUp(self):
-        self.warn_ctx = warnings.catch_warnings(record=True)
-        self.log = self.warn_ctx.__enter__()
-
-        # Do *not* ignore other DeprecationWarnings. Ignoring warnings
-        # can give very confusing results because of
-        # http://bugs.python.org/issue4180 and it is probably simplest to
-        # try to keep the tests cleanly giving only the right warning type.
-        # (While checking them set to "error" those are ignored anyway)
-        # We still have them show up, because otherwise they would be raised
-        warnings.filterwarnings("always", category=DeprecationWarning)
-        warnings.filterwarnings("always", message=self.message,
-                                    category=DeprecationWarning)
-
-    def tearDown(self):
-        self.warn_ctx.__exit__()
-
-    def assert_deprecated(self, function, num=1, ignore_others=False,
-                        function_fails=False,
-                        exceptions=(DeprecationWarning,), args=(), kwargs={}):
         """Test if DeprecationWarnings are given and raised.
 
         This first checks if the function when called gives `num`
@@ -182,6 +81,9 @@ class _DeprecationTestCase(object):
         # reset the log
         self.log[:] = []
 
+        if exceptions is np._NoValue:
+            exceptions = (self.warning_cls,)
+
         try:
             function(*args, **kwargs)
         except (Exception if function_fails else tuple()):
@@ -190,12 +92,12 @@ class _DeprecationTestCase(object):
         # just in case, clear the registry
         num_found = 0
         for warning in self.log:
-            if warning.category is DeprecationWarning:
+            if warning.category is self.warning_cls:
                 num_found += 1
             elif not ignore_others:
                 raise AssertionError(
-                        "expected DeprecationWarning but got: %s" %
-                        (warning.category,))
+                        "expected %s but got: %s" %
+                        (self.warning_cls.__name__, warning.category))
         if num is not None and num_found != num:
             msg = "%i warnings found but %i expected." % (len(self.log), num)
             lst = [str(w.category) for w in self.log]
@@ -203,7 +105,7 @@ class _DeprecationTestCase(object):
 
         with warnings.catch_warnings():
             warnings.filterwarnings("error", message=self.message,
-                                    category=DeprecationWarning)
+                                    category=self.warning_cls)
             try:
                 function(*args, **kwargs)
                 if exceptions != tuple():
@@ -215,7 +117,7 @@ class _DeprecationTestCase(object):
                             "Error raised during function call")
 
     def assert_not_deprecated(self, function, args=(), kwargs={}):
-        """Test if DeprecationWarnings are given and raised.
+        """Test that warnings are not raised.
 
         This is just a shorthand for:
 
@@ -226,40 +128,8 @@ class _DeprecationTestCase(object):
                         exceptions=tuple(), args=args, kwargs=kwargs)
 
 
-class TestBooleanUnaryMinusDeprecation(_DeprecationTestCase):
-    """Test deprecation of unary boolean `-`. While + and * are well
-    defined, unary - is not and even a corrected form seems to have
-    no real uses.
-
-    The deprecation process was started in NumPy 1.9.
-    """
-    message = r"numpy boolean negative, the `-` operator, .*"
-
-    def test_unary_minus_operator_deprecation(self):
-        array = np.array([True])
-        generic = np.bool_(True)
-
-        # Unary minus/negative ufunc:
-        self.assert_deprecated(operator.neg, args=(array,))
-        self.assert_deprecated(operator.neg, args=(generic,))
-
-
-class TestBooleanBinaryMinusDeprecation(_DeprecationTestCase):
-    """Test deprecation of binary boolean `-`. While + and * are well
-    defined, binary  - is not and even a corrected form seems to have
-    no real uses.
-
-    The deprecation process was started in NumPy 1.9.
-    """
-    message = r"numpy boolean subtract, the `-` operator, .*"
-
-    def test_operator_deprecation(self):
-        array = np.array([True])
-        generic = np.bool_(True)
-
-        # Minus operator/subtract ufunc:
-        self.assert_deprecated(operator.sub, args=(array, array))
-        self.assert_deprecated(operator.sub, args=(generic, generic))
+class _VisibleDeprecationTestCase(_DeprecationTestCase):
+    warning_cls = np.VisibleDeprecationWarning
 
 
 class TestRankDeprecation(_DeprecationTestCase):
@@ -307,40 +177,6 @@ class TestComparisonDeprecations(_DeprecationTestCase):
         # following works (and returns False) due to dtype mismatch:
         a == []
 
-    def test_none_comparison(self):
-        # Test comparison of None, which should result in element-wise
-        # comparison in the future. [1, 2] == None should be [False, False].
-        with warnings.catch_warnings():
-            warnings.filterwarnings('always', '', FutureWarning)
-            assert_warns(FutureWarning, operator.eq, np.arange(3), None)
-            assert_warns(FutureWarning, operator.ne, np.arange(3), None)
-
-        with warnings.catch_warnings():
-            warnings.filterwarnings('error', '', FutureWarning)
-            assert_raises(FutureWarning, operator.eq, np.arange(3), None)
-            assert_raises(FutureWarning, operator.ne, np.arange(3), None)
-
-    def test_scalar_none_comparison(self):
-        # Scalars should still just return False and not give a warnings.
-        # The comparisons are flagged by pep8, ignore that.
-        with warnings.catch_warnings(record=True) as w:
-            warnings.filterwarnings('always', '', FutureWarning)
-            assert_(not np.float32(1) == None)
-            assert_(not np.str_('test') == None)
-            # This is dubious (see below):
-            assert_(not np.datetime64('NaT') == None)
-
-            assert_(np.float32(1) != None)
-            assert_(np.str_('test') != None)
-            # This is dubious (see below):
-            assert_(np.datetime64('NaT') != None)
-        assert_(len(w) == 0)
-
-        # For documentation purposes, this is why the datetime is dubious.
-        # At the time of deprecation this was no behaviour change, but
-        # it has to be considered when the deprecations are done.
-        assert_(np.equal(np.datetime64('NaT'), None))
-
     def test_void_dtype_equality_failures(self):
         class NotArray(object):
             def __array__(self):
@@ -393,98 +229,6 @@ class TestComparisonDeprecations(_DeprecationTestCase):
                         assert_warns(DeprecationWarning, f, arg1, arg2)
 
 
-class TestIdentityComparisonDeprecations(_DeprecationTestCase):
-    """This tests the equal and not_equal object ufuncs identity check
-    deprecation. This was due to the usage of PyObject_RichCompareBool.
-
-    This tests that for example for `a = np.array([np.nan], dtype=object)`
-    `a == a` it is warned that False and not `np.nan is np.nan` is returned.
-
-    Should be kept in sync with TestComparisonDeprecations and new tests
-    added when the deprecation is over. Requires only removing of @identity@
-    (and blocks) from the ufunc loops.c.src of the OBJECT comparisons.
-    """
-
-    message = "numpy .* will not check object identity in the future."
-
-    def test_identity_equality_mismatch(self):
-        a = np.array([np.nan], dtype=object)
-
-        with warnings.catch_warnings():
-            warnings.filterwarnings('always', '', FutureWarning)
-            assert_warns(FutureWarning, np.equal, a, a)
-            assert_warns(FutureWarning, np.not_equal, a, a)
-
-        with warnings.catch_warnings():
-            warnings.filterwarnings('error', '', FutureWarning)
-            assert_raises(FutureWarning, np.equal, a, a)
-            assert_raises(FutureWarning, np.not_equal, a, a)
-            # And the other do not warn:
-            with np.errstate(invalid='ignore'):
-                np.less(a, a)
-                np.greater(a, a)
-                np.less_equal(a, a)
-                np.greater_equal(a, a)
-
-    def test_comparison_error(self):
-        class FunkyType(object):
-            def __eq__(self, other):
-                raise TypeError("I won't compare")
-
-            def __ne__(self, other):
-                raise TypeError("I won't compare")
-
-        a = np.array([FunkyType()])
-        self.assert_deprecated(np.equal, args=(a, a))
-        self.assert_deprecated(np.not_equal, args=(a, a))
-
-    def test_bool_error(self):
-        # The comparison result cannot be interpreted as a bool
-        a = np.array([np.array([1, 2, 3]), None], dtype=object)
-        self.assert_deprecated(np.equal, args=(a, a))
-        self.assert_deprecated(np.not_equal, args=(a, a))
-
-
-class TestAlterdotRestoredotDeprecations(_DeprecationTestCase):
-    """The alterdot/restoredot functions are deprecated.
-
-    These functions no longer do anything in numpy 1.10, so
-    they should not be used.
-
-    """
-
-    def test_alterdot_restoredot_deprecation(self):
-        self.assert_deprecated(np.alterdot)
-        self.assert_deprecated(np.restoredot)
-
-
-class TestBooleanIndexShapeMismatchDeprecation():
-    """Tests deprecation for boolean indexing where the boolean array
-    does not match the input array along the given dimensions.
-    """
-    message = r"boolean index did not match indexed array"
-
-    def test_simple(self):
-        arr = np.ones((5, 4, 3))
-        index = np.array([True])
-        #self.assert_deprecated(arr.__getitem__, args=(index,))
-        assert_warns(np.VisibleDeprecationWarning,
-                     arr.__getitem__, index)
-
-        index = np.array([False] * 6)
-        #self.assert_deprecated(arr.__getitem__, args=(index,))
-        assert_warns(np.VisibleDeprecationWarning,
-             arr.__getitem__, index)
-
-        index = np.zeros((4, 4), dtype=bool)
-        #self.assert_deprecated(arr.__getitem__, args=(index,))
-        assert_warns(np.VisibleDeprecationWarning,
-             arr.__getitem__, index)
-        #self.assert_deprecated(arr.__getitem__, args=((slice(None), index),))
-        assert_warns(np.VisibleDeprecationWarning,
-             arr.__getitem__, (slice(None), index))
-
-
 class TestDatetime64Timezone(_DeprecationTestCase):
     """Parsing of datetime64 with timezones deprecated in 1.11.0, because
     datetime64 is now timezone naive rather than UTC only.
@@ -631,6 +375,7 @@ class TestNumericStyleTypecodes(_DeprecationTestCase):
             self.assert_deprecated(np.dtype, exceptions=(TypeError,),
                                    args=(dt,))
 
+
 class TestAccumulateKeepDims(_DeprecationTestCase):
     """
     Deprecate the keepdims argument to np.ufunc.accumulate, which was never used or documented
@@ -655,9 +400,10 @@ class TestTestDeprecated(object):
         test_case_instance.assert_deprecated(foo)
         test_case_instance.tearDown()
 
+
 class TestClassicIntDivision(_DeprecationTestCase):
     """
-    See #7949. Deprecate the numeric-style dtypes with -3 flag in python 2 
+    See #7949. Deprecate the numeric-style dtypes with -3 flag in python 2
     if used for division
     List of data types: http://docs.scipy.org/doc/numpy/user/basics.types.html
     """
@@ -671,11 +417,32 @@ class TestClassicIntDivision(_DeprecationTestCase):
             import operator as op
             dt2 = 'bool_'
             for dt1 in deprecated_types:
-                a = np.array([1,2,3], dtype=dt1)    
-                b = np.array([1,2,3], dtype=dt2)    
-                self.assert_deprecated(op.div, args=(a,b)) 
+                a = np.array([1,2,3], dtype=dt1)
+                b = np.array([1,2,3], dtype=dt2)
+                self.assert_deprecated(op.div, args=(a,b))
                 dt2 = dt1
 
 
+class TestNonNumericConjugate(_DeprecationTestCase):
+    """
+    Deprecate no-op behavior of ndarray.conjugate on non-numeric dtypes,
+    which conflicts with the error behavior of np.conjugate.
+    """
+    def test_conjugate(self):
+        for a in np.array(5), np.array(5j):
+            self.assert_not_deprecated(a.conjugate)
+        for a in (np.array('s'), np.array('2016', 'M'),
+                np.array((1, 2), [('a', int), ('b', int)])):
+            self.assert_deprecated(a.conjugate)
+
+
+class TestNPY_CHAR(_DeprecationTestCase):
+    # 2017-05-03, 1.13.0
+    def test_npy_char_deprecation(self):
+        from numpy.core.multiarray_tests import npy_char_deprecation
+        self.assert_deprecated(npy_char_deprecation)
+        assert_(npy_char_deprecation() == 'S1')
+
+
 if __name__ == "__main__":
     run_module_suite()
